/*
 * Heretic PSP
 * Upstream authors can be found in doc/AUTHORS
 * See docs about original license, the README.txt says that
 * code changes must be GPL'ed
 * PSP Port: Copyright (c) 2005 Gianluigi Tiesi <sherpya@netfarm.it>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <doomdef.h>
#include <SDL/SDL.h>

static int SAMPLECOUNT = 512;
#define NUM_CHANNELS 8
#define SAMPLERATE   11025 /* Hz */

/* The actual lengths of all sound effects. */
int lengths[NUMSFX];

/* The channel step amount... */
unsigned int channelstep[NUM_CHANNELS];
/* ... and a 0.16 bit remainder of last step. */
unsigned int channelstepremainder[NUM_CHANNELS];

/* The channel data pointers, start and end. */
unsigned char *channels[NUM_CHANNELS];
unsigned char *channelsend[NUM_CHANNELS];

int channelstart[NUM_CHANNELS];
int channelhandles[NUM_CHANNELS];
int channelids[NUM_CHANNELS];
int steptable[256];
int vol_lookup[128 * 256];

/* Hardware left and right channel volume lookup. */
int *channelleftvol_lookup[NUM_CHANNELS];
int *channelrightvol_lookup[NUM_CHANNELS];

typedef struct _sound_channel
{
    int is_playing;
    unsigned char *ptr;
    int len;
} sound_channel;

sound_channel sndchans[NUM_CHANNELS];

void *getsfx(char *sfxname, int *len)
{
    unsigned char *sfx, *paddedsfx;
    int size, paddedsize, sfxlump;

    if (W_CheckNumForName(sfxname) == -1)
        sfxlump = W_GetNumForName("gldhit");
    else
        sfxlump = W_GetNumForName(sfxname);

    size = W_LumpLength(sfxlump);
    sfx = (unsigned char *) W_CacheLumpNum(sfxlump, PU_STATIC);

    if (!sfx) return NULL;

    /* Pads the sound effect out to the mixing buffer size.
     * The original realloc would interfere with zone memory.
     */

    paddedsize = ((size - 8 + (SAMPLECOUNT - 1)) / SAMPLECOUNT) * SAMPLECOUNT;

    /* Allocate from zone memory. */
    paddedsfx = (unsigned char *) Z_Malloc(paddedsize + 8, PU_STATIC, 0);

    /* Now copy and pad. */
    memcpy(paddedsfx, sfx, size);
    memset(&paddedsfx[size], 128, (paddedsize + 8) - size);

    /* Remove the cached lump. */
    Z_Free(sfx);

    /* Preserve padded length. */
    *len = paddedsize;

    /* Return allocated padded data. */
    return (void *) (paddedsfx + 8);
}

int addsfx(int sfxid, int volume, int step, int seperation)
{
    static unsigned short handlenums = 0;
    int i, slot, rightvol, leftvol;
    int rc = -1;

    int oldest = gametic;
    int oldestnum = 0;

    /* Loop all channels to find oldest SFX. */
    for (i = 0; (i < NUM_CHANNELS) && (channels[i]); i++)
    {
        if (channelstart[i] < oldest)
        {
            oldestnum = i;
            oldest = channelstart[i];
        }
    }

    if (i == NUM_CHANNELS)
        slot = oldestnum;
    else
        slot = i;

    /* Okay, in the less recent channel,
     *  we will handle the new SFX.
     * Set pointer to raw data.
     */

    channels[slot] = (unsigned char *) S_sfx[sfxid].data;

    /* Set pointer to end of raw data. */
    channelsend[slot] = channels[slot] + lengths[sfxid];

    /* Reset current handle number, limited to 0..100. */
    if (!handlenums)
        handlenums = 100;

    /* Assign current handle number.
     * Preserved so sounds could be stopped (unused).
     */
    channelhandles[slot] = rc = handlenums++;
    channelstep[slot] = step;
    channelstepremainder[slot] = 0;
    channelstart[slot] = gametic;

    /* Separation, that is, orientation/stereo.
     *  range is: 1 - 256
     */
    seperation += 1;

    /* Per left/right channel.
     *  x^2 seperation,
     * adjust volume properly.
     */
    volume *= 8;
    leftvol = volume - ((volume * seperation * seperation) >> 16); /* (256 * 256) */
    seperation = seperation - 257;
    rightvol = volume - ((volume * seperation * seperation) >> 16);

    /* Sanity check, clamp volume. */
    if (rightvol < 0 || rightvol > 127)
        I_Error("rightvol out of bounds");

    if (leftvol < 0 || leftvol > 127)
        I_Error("leftvol out of bounds");

    channelleftvol_lookup[slot] = &vol_lookup[leftvol * 256];
    channelrightvol_lookup[slot] = &vol_lookup[rightvol * 256];

    /* Preserve sound SFX id,
     *  e.g. for avoiding duplicates of chainsaw.
     */
    channelids[slot] = sfxid;

    sndchans[slot].is_playing = 1;
    sndchans[slot].ptr = channels[slot];
    sndchans[slot].len = lengths[sfxid];

    return rc;
}


void I_SetChannels(void)
{
    /* Init internal lookups (raw data, mixing buffer, channels).
     * This function sets up internal lookups used during
     * the mixing process.
     */

    int i, j;
    int *steptablemid = steptable + 128;

    /*
     * Generates volume lookup tables
     *  which also turn the unsigned samples
     *  into signed samples.
     */
    for (i = 0 ; i < 128; i++)
    {
        for (j = 0; j < 256; j++)
            vol_lookup[i * 256 + j] = (i * (j - 128) * 256) / 127;
    }
}

void I_SetSfxVolume(int volume)
{
    snd_SfxVolume = volume;
}

void I_SetMusicVolume(int volume)
{
    snd_MusicVolume = volume;
}

int I_GetSfxLumpNum(sfxinfo_t *sfx)
{
    if (!sfx->name) return 0;
    if (sfx->link) sfx = sfx->link;
    return W_GetNumForName(sfx->name);
}

int I_StartSound(int id, int vol, int sep, int pitch, int priority)
{
    /* UNUSED */
    priority = 0;
    id = addsfx(id, vol, steptable[pitch], sep);
    return id;
}

/* FIXME: not a pointer it's pointless! */
void I_StopSound(int handle)
{
    /* UNUSED. */
    handle = 0;
}

int I_SoundIsPlaying(int handle)
{
    return gametic < handle;
}

/* FIXME: same of above */
void I_UpdateSoundParams(int handle, int vol, int sep, int pitch)
{
    /* UNUSED. */
    handle = vol = sep = pitch = 0;
}

void I_ShutdownSound(void)
{
#ifndef PSP
    SDL_CloseAudio();
#endif
}

/* PSP Audio Sound Callback */
#ifdef PSP
void sound_callback(void* buf, unsigned int length, void *userdata)
#else
/* SDL */
void sound_callback(void *userdata, void *buf, unsigned int length)
#endif
{
    int j, k;
    int l, i;
    int pcm = 0;
    char *p = (char *) buf;

    memset(buf, 0, length * 4);
    l = length * 0.25;

    for(i = 0; i < l; i++)
    {
        pcm = 0;
        for(j = 0; j < NUM_CHANNELS; j++)
        {
            if(sndchans[j].is_playing)
            {
                pcm += ((int) (*(sndchans[j].ptr++)) - 0x80) * 1.5f; /* x 2 */
                sndchans[j].len--;
                if(!sndchans[j].len)
                    sndchans[j].is_playing = 0;
            }
        }
        if(pcm < -0x7f) pcm = -0x7f;
        if(pcm > 0x7f) pcm = 0x7f;
        for(k = 0; k < 16; k++)
            *(p++) = (char) pcm;
    }
}

void I_InitSound(void)
{
    int i;

    for (i = 1; i < NUMSFX; i++)
    {
        /* Alias? Example is the chaingun sound linked to pistol. */
        if (!S_sfx[i].link)
            S_sfx[i].data = getsfx(S_sfx[i].name, &lengths[i]); /* Load data from WAD file. */
        else
        {
            /* Previously loaded already? */
            S_sfx[i].data = S_sfx[i].link->data;
            lengths[i] = lengths[(S_sfx[i].link - S_sfx) / sizeof(sfxinfo_t)];
        }
    }
#ifdef PSP
    pspAudioInit();
    pspAudioSetVolume(0, 0x3000, 0x3000);
    pspAudioSetChannelCallback(0, (void *) &sound_callback, NULL);
#else
    /* SDL Audio */
    SDL_AudioSpec desired, obtained;
    memset(&desired, 0, sizeof(desired));
    memset(&obtained, 0, sizeof(obtained));

    desired.freq = 11025;
    //    desired.format = AUDIO_U16SYS;
    desired.format = AUDIO_U8;
    desired.channels = 2;
    desired.samples = SAMPLECOUNT;
    desired.callback = sound_callback;
    desired.userdata = NULL;

    /* Open the audio device */
    if (SDL_OpenAudio(&desired, &obtained) < 0)
    {
        printf("Couldn't open audio: %s\n", SDL_GetError());
        return;
    }
    SDL_PauseAudio(0);

#endif
}

/* MUSIC API.
 * Still no music done.
 * Remains. Dummies.
 */
void I_InitMusic(void) {}
void I_ShutdownMusic(void) {}

static int looping=0;
static int musicdies=-1;

void I_PlaySong(int handle, int looping)
{
    handle = looping = 0;
    musicdies = gametic + TICRATE * 30;
}

void I_PauseSong(int handle)
{
    handle = 0;
}

void I_ResumeSong(int handle)
{
    handle = 0;
}

void I_StopSong(int handle)
{
    handle = 0;
    looping = 0;
    musicdies = 0;
}

void I_UnRegisterSong(int handle)
{
    handle = 0;
}

int I_RegisterSong(void *data)
{
    data = NULL;
    return 1;
}

/* Is the song playing? */
int I_QrySongPlaying(int handle)
{
    handle = 0;
    return looping || musicdies > gametic;
}

