/*
 * Heretic PSP
 * Upstream authors can be found in doc/AUTHORS
 * See docs about original license, the README.txt says that
 * code changes must be GPL'ed
 * PSP Port: Copyright (c) 2005 Gianluigi Tiesi <sherpya@netfarm.it>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <doomdef.h>

/* TYPES */
typedef struct
{
    char identification[4];     /* should be IWAD */
    int numlumps;
    int infotableofs;
} wadinfo_t;

typedef struct
{
    int filepos;
    int size;
    char name[8];
} filelump_t;


/* GLoBALS */
lumpinfo_t *lumpinfo;           /* location of each lump on disk */
int numlumps;

void **lumpcache;

#ifndef PSP
void strupr (char *s)
{
    while (*s)
    {
        *s = toupper(*s);
        s++;
    }
}
#endif

/* filelength */
int filelength(int handle)
{
    struct stat fileinfo;

    if (fstat(handle, &fileinfo) == -1)
        I_Error("Error fstating");

    return fileinfo.st_size;
}

void ExtractFileBase(char *path, char *dest)
{
    char *src;
    int length;

    src = path + strlen(path) - 1;

    /* back up until a \ or the start */
    while (src != path && *(src - 1) != '\\' && *(src - 1) != '/')
        src--;

    /* copy up to eight characters */

    memset(dest, 0, 8);
    length = 0;

    while (*src && *src != '.')
    {
        if (++length == 9)
            I_Error("Filename base of %s >8 chars", path);
        *dest++ = toupper((int) *src++);
    }
}


/* Helper function */
int wadopen(const char *fileName)
{
    int fd = -1;
    char *fn;

    fn = (char *) malloc(strlen("./") + strlen(fileName) + 1);
    assert(fn);
    sprintf(fn, "./%s", fileName);
    if (!access(fn, R_OK))
    {
        fd = open(fn, O_RDONLY);
        free(fn);
        return fd;
    }
    return fd;
}

/* LUMP BASED ROUTINES */

/* W_AddFile:
 *  All files are optional, but at least one file must be found
 *  Files with a .wad extension are wadlink files with multiple lumps
 *  Other files are single lumps with the base filename for the lump name
 */
void W_AddFile(char *filename)
{
    wadinfo_t header;
    lumpinfo_t *lump_p;
    unsigned i;
    int handle;
    size_t length;
    int startlump;
    filelump_t *fileinfo;
    /* 
     * open the file and add to directory
     */

    if ((handle = wadopen(filename)) == -1)
    {
        printf(" couldn't open %s\n", filename);
        return;
    }

    printf(" adding %s\n", filename);
    startlump = numlumps;

    /* WAD file */
    read(handle, &header, sizeof(header));
    if (strncmp(header.identification, "IWAD", 4))
    {
        if (strncmp(header.identification, "PWAD", 4))
            I_Error("Wad file %s doesn't have IWAD or PWAD id\n", filename);
    }
    header.numlumps = header.numlumps;
    header.infotableofs = header.infotableofs;
    length = header.numlumps * sizeof(filelump_t);
    fileinfo = alloca(length);
    lseek(handle, header.infotableofs, SEEK_SET);
    read(handle, fileinfo, length);
    numlumps += header.numlumps;

    /* Fill in lumpinfo */
    lumpinfo = realloc(lumpinfo, numlumps * sizeof(lumpinfo_t));
    if (!lumpinfo)
        I_Error("Couldn't allocate lumpinfo");
    lump_p = &lumpinfo[startlump];

    for (i = startlump; i < numlumps; i++, lump_p++, fileinfo++)
    {
        lump_p->handle = handle;
        lump_p->position = fileinfo->filepos;
        lump_p->size = fileinfo->size;
        strncpy(lump_p->name, fileinfo->name, 8);
    }
}


/* W_InitMultipleFiles:
 *  Pass a null terminated list of files to use.
 *  All files are optional, but at least one file must be found
 *  Files with a .wad extension are idlink files with multiple lumps
 *  Other files are single lumps with the base filename for the lump name
 *  Lump names can appear multiple times. The name searcher looks backwards,
 *  so a later file can override an earlier one.
 */
void W_InitMultipleFiles(char **filenames)
{
    int size;

    /* open all the files, load headers, and count lumps */
    numlumps = 0;
    lumpinfo = malloc(1); /* will be realloced as lumps are added */

    for (; *filenames; filenames++)
        W_AddFile(*filenames);

    if (!numlumps)
        I_Error("W_InitFiles: no files found");

    /* set up caching */
    size = numlumps * sizeof(*lumpcache);
    lumpcache = malloc(size);
    assert(lumpcache);
    if (!lumpcache)
        I_Error("Couldn't allocate lumpcache");
    memset(lumpcache, 0, size);
}


/* W_InitFile:
 *  Just initialize from a single file
 */
void W_InitFile(char *filename)
{
    char *names[2];

    names[0] = filename;
    names[1] = NULL;
    W_InitMultipleFiles(names);
}


/* W_NumLumps */
int W_NumLumps(void)
{
    return numlumps;
}

/* W_CheckNumForName:
 *  Returns -1 if name not found
 */
int W_CheckNumForName(char *name)
{
    char name8[9];
    int v1, v2;
    lumpinfo_t *lump_p;

    /* make the name into two integers for easy compares */

    strncpy(name8, name, 8);
    name8[8] = 0; /* in case the name was a fill 8 chars */
    strupr(name8); /* case insensitive */

    v1 = *(int *) name8;
    v2 = *(int *) &name8[4];

    /* scan backwards so patch lump files take precedence */

    lump_p = lumpinfo + numlumps;

    while (lump_p-- != lumpinfo)
    {
        if (*(int *) lump_p->name == v1 && *(int *) &lump_p->name[4] == v2)
            return lump_p - lumpinfo;
    }
    return -1;
}


/* W_GetNumForName:
 *  Calls W_CheckNumForName, but bombs out if not found
 */
int W_GetNumForName(char *name)
{
    int i;

    i = W_CheckNumForName(name);
    if (i != -1)
        return i;

    I_Error("W_GetNumForName: %s not found!", name);
    return -1; /* Never reached */
}


/* W_LumpLength:
 *  Returns the buffer size needed to load the given lump
 */
int W_LumpLength(int lump)
{
    if (lump >= numlumps)
        I_Error("W_LumpLength: %i >= numlumps", lump);
    return lumpinfo[lump].size;
}


/* W_ReadLump:
 *  Loads the lump into the given buffer, which must be >= W_LumpLength()
 */
void W_ReadLump(int lump, void *dest)
{
    int c;
    lumpinfo_t *l;

    if (lump >= numlumps)
        I_Error("W_ReadLump: %i >= numlumps", lump);
    l = lumpinfo + lump;

    /* I_BeginRead (); - for use with DiskIconFlashing... */

    lseek(l->handle, l->position, SEEK_SET);
    c = read(l->handle, dest, l->size);

    if (c < l->size)
        I_Error("W_ReadLump: only read %i of %i on lump %i", c, l->size, lump);
    /* I_EndRead (); - for use with DiskIconFlashing... */
}

/* W_CacheLumpNum */
void *W_CacheLumpNum(int lump, int tag)
{
    byte *ptr;

    if ((unsigned) lump >= numlumps)
        I_Error("W_CacheLumpNum: %i >= numlumps", lump);

    if (!lumpcache[lump])
    {
        /* read the lump in */
        /* printf ("cache miss on lump %i\n",lump); */
        ptr = Z_Malloc(W_LumpLength(lump), tag, &lumpcache[lump]);
        W_ReadLump(lump, lumpcache[lump]);
    }
    else
        Z_ChangeTag(lumpcache[lump], tag); /* printf ("cache hit on lump %i\n",lump); */

    return lumpcache[lump];
}


/* W_CacheLumpName */
void *W_CacheLumpName(char *name, int tag)
{
    return W_CacheLumpNum(W_GetNumForName(name), tag);
}
